//2002.3.2
//O.Manba

module USB2LCD(
		CLK,
		RST_X,
		D,
		RD_X,
		WR,
		RXF_X,
		TXE_X,
		DB,
		EEE,
		RWB,
		RS
	);
input			CLK;
input			RST_X;
inout	[7:0]	D;
output			RD_X;
output			WR;
input			RXF_X;
input			TXE_X;
output	[7:0]	DB;
output			EEE;
output			RWB;
output			RS;

reg		[1:0]	STATE_REG;
reg		[7:0]	DATA_REG;
reg				WR_REG;
reg				RD_X_REG;
reg				TXE_X_REG;
reg				RXF_X_REG;
reg				SEND_REQ_REG;

wire	[7:0]	D_PRE;
wire	[7:0]	DBO;
wire			SEND_ACK;

assign D_PRE[7:0] = (RD_X_REG == 1'b1) ? DATA_REG[7:0] : 8'b1111_1111;

assign D[7] = (D_PRE[7] == 1'b0) ? 1'b0 : 1'bz;
assign D[6] = (D_PRE[6] == 1'b0) ? 1'b0 : 1'bz;
assign D[5] = (D_PRE[5] == 1'b0) ? 1'b0 : 1'bz;
assign D[4] = (D_PRE[4] == 1'b0) ? 1'b0 : 1'bz;
assign D[3] = (D_PRE[3] == 1'b0) ? 1'b0 : 1'bz;
assign D[2] = (D_PRE[2] == 1'b0) ? 1'b0 : 1'bz;
assign D[1] = (D_PRE[1] == 1'b0) ? 1'b0 : 1'bz;
assign D[0] = (D_PRE[0] == 1'b0) ? 1'b0 : 1'bz;
assign RD_X = (RD_X_REG == 1'b0) ? 1'b0 : 1'bz;
assign WR   = (WR_REG   == 1'b0) ? 1'b0 : 1'bz;

assign DB[7] = DBO[7];
assign DB[6] = DBO[6];
assign DB[5] = DBO[5];
assign DB[4] = DBO[4];
assign DB[3] = DBO[3];
assign DB[2] = DBO[2];
assign DB[1] = DBO[1];
assign DB[0] = DBO[0];

LCD_DRIVER lcd_driver(
		.CLK		(CLK),
		.RST_X		(RST_X),
		.DBO		(DBO[7:0]),
		.EEE		(EEE),
		.RWB		(RWB),
		.RS			(RS),
		.SEND_REQ	(SEND_REQ_REG),
		.SEND_ACK	(SEND_ACK),
		.SEND_DATA	(DATA_REG[7:0])
	);

always @(posedge CLK or negedge RST_X) begin
	if (RST_X == 1'b0) begin
		TXE_X_REG <= 1'b1;
		RXF_X_REG <= 1'b1;
	end
	else begin
		TXE_X_REG <= TXE_X;
		RXF_X_REG <= RXF_X;
	end
end

always @(posedge CLK or negedge RST_X) begin
	if (RST_X == 1'b0) begin
		STATE_REG[1:0] <= 2'b00;
		DATA_REG[7:0] <= 8'h00;
		WR_REG <= 1'b1;
		RD_X_REG <= 1'b1;
		SEND_REQ_REG <= 1'b0;
	end
	else begin
		if (STATE_REG[1:0] == 2'b00) begin
			if (RD_X_REG == 1'b0) begin
				STATE_REG <= 2'b01;
				DATA_REG[7:0] <= D[7:0];
				RD_X_REG <= 1'b1;
			end
			else if (RXF_X_REG == 1'b0) begin
				RD_X_REG <= 1'b0;
			end
		end
		else if (STATE_REG[1:0] == 2'b01) begin
			if (WR_REG == 1'b0) begin
				STATE_REG[1:0] <= 2'b10;
				WR_REG <= 1'b1;
			end
			else if (TXE_X_REG == 1'b0) begin
				WR_REG <= 1'b0;
			end
		end
		else if (STATE_REG[1:0] == 2'b10) begin
			STATE_REG[1:0] <= 2'b11;
			SEND_REQ_REG <= 1'b1;
		end
		else begin
			if (SEND_ACK == 1'b1) begin
				STATE_REG[1:0] <= 2'b00;
				SEND_REQ_REG <= 1'b0;
			end
		end
	end
end

endmodule

module LCD_DRIVER(
		CLK,
		RST_X,
		DBO,
		EEE,
		RWB,
		RS,
		SEND_REQ,
		SEND_ACK,
		SEND_DATA
	);
input			CLK;
input			RST_X;
output	[7:0]	DBO;
output			EEE;
output			RWB;
output			RS;
input			SEND_REQ;
output			SEND_ACK;
input	[7:0]	SEND_DATA;

reg		[4:0]	STATE;
reg		[7:0]	DBO_REG;
reg				EEE_REG;
reg				RS_REG;
reg				SEND_ACK_REG;
reg		[4:0]	LINE_COUNT_REG;
reg		[4:0]	LINE_COUNT_REG_POST;
reg		[7:0]	SEND_DATA_REG;

assign DBO[7]   = DBO_REG[7];
assign DBO[6]   = DBO_REG[6];
assign DBO[5]   = DBO_REG[5];
assign DBO[4]   = DBO_REG[4];
assign DBO[3]   = DBO_REG[3];
assign DBO[2]   = DBO_REG[2];
assign DBO[1]   = DBO_REG[1];
assign DBO[0]   = DBO_REG[0];
assign EEE      = EEE_REG;
assign RWB      = 1'b0;
assign RS       = RS_REG;
assign SEND_ACK = SEND_ACK_REG;

always @(posedge CLK or negedge RST_X) begin
	if (RST_X == 1'b0) begin
		STATE[4:0] <= 5'd0;
		DBO_REG[7:0] <= 8'b00000000;
		EEE_REG <= 1'b0;
		RS_REG <= 1'b0;
		SEND_ACK_REG <= 1'b0;
		LINE_COUNT_REG[4:0] <= 5'h00;
		SEND_DATA_REG[7:0] <= 8'h00;
	end
	else begin
		case(STATE[4:0])
//
// initialize
//
		5'd0: begin
			STATE[4:0] <= 5'd1;
			DBO_REG[7:0] <= 8'b00000000;
			EEE_REG <= 1'b0;
			RS_REG  <= 1'b0;
			SEND_ACK_REG <= 1'b0;
		end
//
// function set
//
		5'd1: begin
			STATE[4:0] <= 5'd2;
			DBO_REG[7:0] <= 8'b00111000;
			EEE_REG <= 1'b0;
			RS_REG  <= 1'b0;
		end
		5'd2: begin
			STATE[4:0] <= 5'd3;
			EEE_REG <= 1'b1;
			RS_REG  <= 1'b0;
		end
		5'd3: begin
			STATE[4:0] <= 5'd4;
			EEE_REG <= 1'b0;
			RS_REG  <= 1'b0;
		end
//
// display on/off
//
		5'd4: begin
			STATE[4:0] <= 5'd5;
			DBO_REG[7:0] <= 8'b00001110;
			EEE_REG <= 1'b0;
			RS_REG  <= 1'b0;
		end
		5'd5: begin
			STATE[4:0] <= 5'd6;
			EEE_REG <= 1'b1;
			RS_REG  <= 1'b0;
		end
		5'd6: begin
			STATE[4:0] <= 5'd7;
			EEE_REG <= 1'b0;
			RS_REG  <= 1'b0;
		end
//
// entry mode set
//
		5'd7: begin
			STATE[4:0] <= 5'd8;
			DBO_REG[7:0] <= 8'b00000110;
			EEE_REG <= 1'b0;
			RS_REG  <= 1'b0;
		end
		5'd8: begin
			STATE[4:0] <= 5'd9;
			EEE_REG <= 1'b1;
			RS_REG  <= 1'b0;
		end
		5'd9: begin
			STATE[4:0] <= 5'd10;
			EEE_REG <= 1'b0;
			RS_REG  <= 1'b0;
		end
//
// display clear
//
		5'd10: begin
			STATE[4:0] <= 5'd11;
			DBO_REG[7:0] <= 8'b00000001;
			EEE_REG <= 1'b0;
			RS_REG  <= 1'b0;
		end
		5'd11: begin
			STATE[4:0] <= 5'd12;
			EEE_REG <= 1'b1;
			RS_REG  <= 1'b0;
		end
		5'd12: begin
			STATE[4:0] <= 5'd13;
			EEE_REG <= 1'b0;
			RS_REG  <= 1'b0;
		end
//
// main loop start
//
		5'd13: begin
			EEE_REG <= 1'b0;
			RS_REG  <= 1'b0;
			SEND_ACK_REG <= 1'b0;
			
			if (SEND_REQ == 1'b1) begin
				// Back Space (08H)
				if (SEND_DATA[7:0] == 8'h08) begin
					STATE[4:0] <= 5'd14;
					SEND_DATA_REG[7:0] <= 8'h20;
					if (LINE_COUNT_REG[3:0] != 4'd0) begin
						LINE_COUNT_REG_POST[4:0] <= LINE_COUNT_REG[4:0] - 5'd1;
						LINE_COUNT_REG[4:0]      <= LINE_COUNT_REG[4:0] - 5'd1;
					end
				end
				// Entar (0DH)
				else if (SEND_DATA[7:0] == 8'h0d) begin
					STATE[4:0] <= 5'd20;
					SEND_DATA_REG[7:0] <= 8'h20;
					LINE_COUNT_REG_POST[4:0] <= {~LINE_COUNT_REG[4], 4'd0};
					LINE_COUNT_REG[4:0]      <= LINE_COUNT_REG[4:0];
				end
				// Others
				else begin
					STATE[4:0] <= 5'd14;
					SEND_DATA_REG[7:0] <= SEND_DATA[7:0];
					LINE_COUNT_REG_POST[4:0] <= LINE_COUNT_REG[4:0] + 5'd1;
					LINE_COUNT_REG[4:0]      <= LINE_COUNT_REG[4:0];
				end
			end
			else begin
				STATE[4:0] <= 5'd13;
			end
		end
//
// address set (before)
//
		5'd14: begin
			STATE[4:0] <= 5'd15;
			DBO_REG[7:0] <= {1'b1, LINE_COUNT_REG[4], 2'b00, LINE_COUNT_REG[3:0]};
			EEE_REG <= 1'b0;
			RS_REG  <= 1'b0;
		end
		5'd15: begin
			STATE[4:0] <= 5'd16;
			EEE_REG <= 1'b1;
			RS_REG  <= 1'b0;
		end
		5'd16: begin
			STATE[4:0] <= 5'd17;
			EEE_REG <= 1'b0;
			RS_REG  <= 1'b0;
		end
//
// write DD RAM
//
		5'd17: begin
			STATE[4:0] <= 5'd18;
			DBO_REG[7:0] <= SEND_DATA_REG[7:0];
			EEE_REG <= 1'b0;
			RS_REG  <= 1'b1;
		end
		5'd18: begin
			STATE[4:0] <= 5'd19;
			EEE_REG <= 1'b1;
			RS_REG  <= 1'b1;
		end
		5'd19: begin
			STATE[4:0] <= 5'd20;
			EEE_REG <= 1'b0;
			RS_REG  <= 1'b1;
		end
//
// send ack
//
		5'd20: begin
			STATE[4:0] <= 5'd21;
			EEE_REG <= 1'b0;
			RS_REG  <= 1'b0;
			SEND_ACK_REG <= 1'b1;
			LINE_COUNT_REG[4:0] <= LINE_COUNT_REG_POST[4:0];
		end
		5'd21: begin
			STATE[4:0] <= 5'd22;
			EEE_REG <= 1'b0;
			RS_REG  <= 1'b0;
			SEND_ACK_REG <= 1'b0;
		end
//
// address set (after)
//
		5'd22: begin
			STATE[4:0] <= 5'd23;
			DBO_REG[7:0] <= {1'b1, LINE_COUNT_REG[4], 2'b00, LINE_COUNT_REG[3:0]};
			EEE_REG <= 1'b0;
			RS_REG  <= 1'b0;
		end
		5'd23: begin
			STATE[4:0] <= 5'd24;
			EEE_REG <= 1'b1;
			RS_REG  <= 1'b0;
		end
		5'd24: begin
			STATE[4:0] <= 5'd25;
			EEE_REG <= 1'b0;
			RS_REG  <= 1'b0;
		end
//
// main loop end
//
		5'd25: begin
			STATE[4:0] <= 5'd13;
			EEE_REG <= 1'b0;
			RS_REG  <= 1'b0;
		end
//
// default
//
		default: begin
			STATE[4:0] <= 5'd0;
			DBO_REG[7:0] <= 8'b00000000;
			EEE_REG <= 1'b0;
			RS_REG  <= 1'b0;
			SEND_ACK_REG <= 1'b0;
		end
		endcase
	end
end

endmodule
